class Direction:
    north_east=1
    east=2
    south_east=3
    south=4
    south_west=-1
    west=-2
    north_west=-3
    north= -4

    def __init__(self):
        self.north_east = 1
        self.east = 2
        self.south_east = 3
        self.south = 4
        self.south_west = -1
        self.west = -2
        self.north_west = -3
        self.north = -4
